function [INP2HID, HID2OUT, ITER] = ann_calc_weights(INPUT_DIM, HID_LAYER_DIM, OUTPUT_DIM, FEATURE_MATRIX_TRAINING, USE_MOMENTUM)
%ANN_CALC_WEIGHTS - uses the well-known backpropagation algorithm to train the weights of a
% three-layer artificial network

%% INITIALIZE VARIABLES
TRAINING_LENGTH = size(FEATURE_MATRIX_TRAINING,1);

% initialize weights (uniformly distributed)
INP2HID = 2*(rand(HID_LAYER_DIM, INPUT_DIM + 1) - 0.5);
HID2OUT = 2*(rand(OUTPUT_DIM, HID_LAYER_DIM + 1)- 0.5);

% momentum term
alpha = 0.9;

OLD_UPDATE_HID2OUT = zeros(OUTPUT_DIM, HID_LAYER_DIM + 1);
OLD_UPDATE_INP2HID = zeros(HID_LAYER_DIM, INPUT_DIM + 1);

TARGET = repmat(0.1, OUTPUT_DIM, TRAINING_LENGTH);
TARGET(sub2ind(size(TARGET), FEATURE_MATRIX_TRAINING(:, end)', 1:TRAINING_LENGTH)) = 0.9;

INPUT           = [FEATURE_MATRIX_TRAINING(:, 1:INPUT_DIM)'; ones(1,TRAINING_LENGTH)];

%% ITERATIVELY UPDATE WEIGHTS OF NEURAL NETWORK
ITER = 0;
while (ITER < 5 || std(J(ITER-4:ITER)) > 0.5e-3)
    ITER = ITER + 1;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % calculate J(ITER)
    J(ITER) = norm(ann_calc_output(INP2HID, HID2OUT, INPUT(1:end-1,:)) - TARGET, 'fro')^2;
    J(ITER) = J(ITER) / TRAINING_LENGTH;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    RAND_INDICES = randperm(TRAINING_LENGTH);

    % step-size
    eta = 1/sqrt(ITER);

    for ROW = 1:TRAINING_LENGTH
        % calculate output
        HID_LAYER_IN    = INP2HID * INPUT(:,RAND_INDICES(ROW)); % the 1 is there because we also allow a bias
        [HID_LAYER_OUT, dHID_LAYER_OUT] = sigmoid(HID_LAYER_IN);

        OUTPUT_LAYER_IN   = HID2OUT * [HID_LAYER_OUT; 1];
        [OUTPUT, dOUTPUT] = sigmoid(OUTPUT_LAYER_IN);

        % calculate sensitivity for each output node
        DELTA_HID2OUT = (TARGET(:,RAND_INDICES(ROW)) - OUTPUT) .* dOUTPUT;

        % calculate sensitivity for each hidden node
        TMP = HID2OUT(:,1:end-1)' * DELTA_HID2OUT;
        DELTA_INP2HID = dHID_LAYER_OUT .* TMP;

        % Update weights
        BP_UPDATE_HID2OUT = eta * DELTA_HID2OUT * [HID_LAYER_OUT' 1];
        BP_UPDATE_INP2HID = eta * DELTA_INP2HID * INPUT(:,RAND_INDICES(ROW))';

        if USE_MOMENTUM == 0
            HID2OUT = HID2OUT + BP_UPDATE_HID2OUT;
            INP2HID = INP2HID + BP_UPDATE_INP2HID;
        else
            % do BP step in first iteration
            if ITER == 1 && ROW == 1
                OLD_UPDATE_HID2OUT = BP_UPDATE_HID2OUT;
                OLD_UPDATE_INP2HID = BP_UPDATE_INP2HID;
            else
                OLD_UPDATE_HID2OUT = (1-alpha) * BP_UPDATE_HID2OUT + alpha * OLD_UPDATE_HID2OUT;
                OLD_UPDATE_INP2HID = (1-alpha) * BP_UPDATE_INP2HID + alpha * OLD_UPDATE_INP2HID;
            end
            HID2OUT = HID2OUT + OLD_UPDATE_HID2OUT;
            INP2HID = INP2HID + OLD_UPDATE_INP2HID;
        end
    end

    fprintf('%d\t%f\n', ITER, J(ITER));
end

%% SUBFUNCTION sigmoid
function [Y, dY] = sigmoid(X)
% sigmoidal activation function
% see Haykin: Adaptive Filter Theory, pp. 736 - 760
Y  = 1./(1 + exp(-X));
dY = Y.*(1 - Y);  % differentiated sigmoidal function