function [INP2HID, HID2OUT, ITER] = ann_calc_weights_hess(INPUT_DIM, HID_LAYER_DIM, OUTPUT_DIM, FEATURE_MATRIX_TRAINING)
%ANN_CALC_WEIGHTS - uses the well-known Levenberg-Marquardt algorithm
% to train the weights of a three-layer artificial network
% The Levenberg-Marquardt algorithm is based on an outer-product approximation of the hessian matrix
% see Hagan (1994): Training Feedforward Networks with the Marquardt algorithm

%% INITIALIZE VARIABLES
% initialize weights (uniformly distributed)
INP2HID = 2*(rand(HID_LAYER_DIM, INPUT_DIM + 1)- 0.5);
HID2OUT = 2*(rand(OUTPUT_DIM, HID_LAYER_DIM + 1)-0.5);

TRAINING_LENGTH = size(FEATURE_MATRIX_TRAINING,1);
NUM_UNKNOWNS_INP2HID = numel(INP2HID);
NUM_UNKNOWNS_HID2OUT = numel(HID2OUT);
NUM_UNKNOWNS         = NUM_UNKNOWNS_INP2HID + NUM_UNKNOWNS_HID2OUT;

% TARGET contains the desired outputs of the ANN (OUTPUT_DIM x TRAINING_LENGTH)
TARGET = repmat(0.1, OUTPUT_DIM, TRAINING_LENGTH);
TARGET(sub2ind(size(TARGET), FEATURE_MATRIX_TRAINING(:, end)', 1:TRAINING_LENGTH)) = 0.9;
TARGET_VEC = TARGET(:);
% INDICES_HID2OUT and INDICES_INP2HID are used to transform the outer product of two vectors into the hadamard product, i.e. .*
% This is done to vectorize the calculation of the hessian matrix
INDICES_HID2OUT = repmat(1:(HID_LAYER_DIM+1)*TRAINING_LENGTH*OUTPUT_DIM,OUTPUT_DIM,1);
INDICES_HID2OUT = reshape(INDICES_HID2OUT,NUM_UNKNOWNS_HID2OUT,numel(INDICES_HID2OUT)/NUM_UNKNOWNS_HID2OUT);
INDICES_INP2HID = repmat(1:(INPUT_DIM+1)*TRAINING_LENGTH*OUTPUT_DIM,HID_LAYER_DIM,1);
INDICES_INP2HID = reshape(INDICES_INP2HID,NUM_UNKNOWNS_INP2HID,numel(INDICES_INP2HID)/NUM_UNKNOWNS_INP2HID);

% INPUT contains the augmented input features
INPUT           = [FEATURE_MATRIX_TRAINING(:, 1:INPUT_DIM)'; ones(1,TRAINING_LENGTH)];

% INPUT_EXPANDED is needed to calculate DELTA_HID2OUT. We could also use INPUT and then expand dOUTPUT as it is
% shown in equation (13) in the paper of Hagan. However the computational savings are not that much and it is easier to use
% INPUT_EXPANDED
INPUT_EXPANDED  = reshape(repmat(INPUT,OUTPUT_DIM,1),INPUT_DIM+1,TRAINING_LENGTH*OUTPUT_DIM);
INPUT_LAYER_VEC = INPUT_EXPANDED(INDICES_INP2HID);

% INDICES_EXPANDED... indices to obtain DELTA_HID2OUT out of dOUTPUT (corresponds to eq. 13 of paper from Hagan)
INDICES_EXPANDED = [];
for ROW = 1:OUTPUT_DIM
    INDICES_EXPANDED = [INDICES_EXPANDED; ([1:OUTPUT_DIM^2:OUTPUT_DIM^2*TRAINING_LENGTH]+(ROW-1)*(OUTPUT_DIM+1))'];
end

% INDICES_EXPANDED2... indices to obtain the output vectors out of the expanded output vector
INDICES_EXPANDED2 = [];
for ROW = 1:TRAINING_LENGTH
    INDICES_EXPANDED2 = [INDICES_EXPANDED2; ([1:OUTPUT_DIM]+(ROW-1)*(OUTPUT_DIM)^2)'];
end

% define LAMBDA (see paper by Marquardt, "An algorithm for least-squares estimation of nonlinear parameters"
LAMBDA = 0.01;
NU     = 10; % step parameter for switching between newton method and gradient descent

% calculate error before first update
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% calculate CRIT_J
CRIT_J1 = norm(ann_calc_output(INP2HID, HID2OUT, INPUT(1:end-1,:)) - TARGET, 'fro')^2;
CRIT_J1 = CRIT_J1 / TRAINING_LENGTH;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% ITERATIVELY UPDATE WEIGHTS OF NEURAL NETWORK (USE NEWTON-METHOD)
ITER = 0;
while (ITER<5 || std(J(ITER-4:ITER)) > 1e-5)

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % calculate Hessian matrix by the outer product approximation

    % calculate output (Forward Propagation)
    HID_LAYER_IN    = INP2HID * INPUT_EXPANDED;
    [HID_LAYER_OUT, dHID_LAYER_OUT] = sigmoid(HID_LAYER_IN);

    OUTPUT_LAYER_IN   = HID2OUT * [HID_LAYER_OUT; ones(1, size(HID_LAYER_OUT,2))];
    [OUTPUT, dOUTPUT] = sigmoid(OUTPUT_LAYER_IN);

    % calculate sensitivity for each output node (Backward Propagation)
    % Note, that (TARGET - OUTPUT) is neglected in comparison to the standard BP algorithm
    DELTA_HID2OUT = zeros(OUTPUT_DIM,TRAINING_LENGTH*OUTPUT_DIM);
    DELTA_HID2OUT(INDICES_EXPANDED) = dOUTPUT(INDICES_EXPANDED);
    
    % calculate sensitivity for each hidden node
    TMP = DELTA_HID2OUT' * HID2OUT;
    DELTA_INP2HID = dHID_LAYER_OUT .* TMP(:,1:end-1)';

    BP_UPDATE_HID2OUT = repmat(DELTA_HID2OUT,HID_LAYER_DIM+1,1);
    TMP               = [HID_LAYER_OUT; ones(1,size(HID_LAYER_OUT,2))];
    HIDDEN_LAYER_VEC  = TMP(INDICES_HID2OUT);
    BP_UPDATE_INP2HID = repmat(DELTA_INP2HID,INPUT_DIM+1,1);

    VEC = [BP_UPDATE_INP2HID.*INPUT_LAYER_VEC; BP_UPDATE_HID2OUT.*HIDDEN_LAYER_VEC];

    HESSIAN     = VEC*VEC';
    HESS_MATRIX = HESSIAN + LAMBDA*eye(NUM_UNKNOWNS);

    BATCH_GRAD  = VEC * (TARGET_VEC - OUTPUT(INDICES_EXPANDED2));

    while rcond(HESS_MATRIX) < 1e-8
        %fprintf('Hessian matrix singular. Increasing LAMBDA (%f)\n',LAMBDA);
        LAMBDA = LAMBDA * NU;
        HESS_MATRIX = HESSIAN + LAMBDA*eye(NUM_UNKNOWNS);
    end

    % if 1 then weights were updated and Hessian matrix has to computed again
    UPDATED = 0;
    
    while UPDATED == 0

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Update values for weights

    UPDATE_VEC = HESS_MATRIX \ BATCH_GRAD;
    UPDATE_INP2HID = reshape(UPDATE_VEC(1:NUM_UNKNOWNS_INP2HID),size(INP2HID));
    UPDATE_HID2OUT = reshape(UPDATE_VEC(NUM_UNKNOWNS_INP2HID+1:end),size(HID2OUT));

    % calculate error after update
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % calculate CRIT_J
    CRIT_J2 = norm(ann_calc_output(INP2HID+UPDATE_INP2HID, HID2OUT+UPDATE_HID2OUT, INPUT(1:end-1,:)) - TARGET, 'fro')^2;
    CRIT_J2 = CRIT_J2 / TRAINING_LENGTH;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if CRIT_J2 < CRIT_J1
        % step has decreased errror -> accept step and update weights
        % LAMBDA is decreased
        LAMBDA = LAMBDA / NU;
        HID2OUT = HID2OUT + UPDATE_HID2OUT;
        INP2HID = INP2HID + UPDATE_INP2HID;
        CRIT_J1 = CRIT_J2;        
        UPDATED = 1;
    else
        if LAMBDA < 1e10
            LAMBDA = LAMBDA * NU;
            HESS_MATRIX = HESSIAN + LAMBDA*eye(NUM_UNKNOWNS);
        end
    end
    end
    ITER = ITER + 1;
    J(ITER) = CRIT_J1;
%    fprintf('%d\tJ: %f LAMBDA: %e\n', ITER, CRIT_J1,LAMBDA);
end

%% SUBFUNCTION sigmoid
function [Y, dY] = sigmoid(X)
% sigmoidal activation function
% see Haykin: Adaptive Filter Theory, pp. 736 - 760
Y  = 1./(1 + exp(-X));
dY = Y.*(1 - Y);  % differentiated sigmoidal function