function EVAL_MATRIX = ann_classifier(DATA, PARAMETERS)
%ANN_CLASSIFIER - uses an artificial neural network to classify data
% we classify ITERATIONS times to obtain a stable value
% Possible parameters are:
%
% PARAMETERS.ANN_ITERATIONS      - number of iterations
% PARAMETERS.USE_MARQUARDT       - = 0: use statistical annealing aka Backpropagation
%                                       algorithm. 
%                                  = 1: use Levenberg-Marquardt algorithm which uses
%                                       a Gauss-Newton method
% PARAMETERS.HIDDEN_LAYER_LENGTH - number of hidden layer nodes
% PARAMETERS.USE_MOMENTUM_TERM   - = 1: then we will use momentum to speed up the
%                                       backpropagation algorithm
%                                  = 0: use simple backprop. algorithm

%% 1. INITIALISATION
[TRAINING_LENGTH, CLASSIFICATION_LENGTH, NUM_CLASSES, CLASSES, FEATURE_LENGTH] = get_lengths(DATA, PARAMETERS);

% number of total iterations to get stable results
if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'ANN_ITERATIONS') == 0
    PARAMETERS.ANN_ITERATIONS = 10;
end

% number of hidden layers
if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'HIDDEN_LAYER_LENGTH') == 0
    PARAMETERS.HIDDEN_LAYER_LENGTH = 3;
end

% number of hidden layers
if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'USE_MOMENTUM_TERM') == 0
    PARAMETERS.USE_MOMENTUM_TERM = 1;
end

% use standard BP or LM algorithm
if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'USE_MARQUARDT') == 0
    PARAMETERS.USE_MARQUARDT = 0;
end

% normalize input
NORMALIZED = normalize_matrix(DATA);

% transform class labels to indices which are used in CLASSES
for CLASSES_C = 1:NUM_CLASSES
    INDICES{CLASSES_C} = NORMALIZED.TRAINING(:,end) == CLASSES(CLASSES_C);
end
for CLASSES_C = 1:NUM_CLASSES
    NORMALIZED.TRAINING(INDICES{CLASSES_C},end) = CLASSES_C;
end

%% 2. TRAININGSPHASE
ITER_M = zeros(PARAMETERS.ANN_ITERATIONS, 1);
P_CORRECT = 0;
for ITER = 1:PARAMETERS.ANN_ITERATIONS

    % 1. calculate weights
    if PARAMETERS.USE_MARQUARDT == 0
        [INP2HID_NEW, HID2OUT_NEW, ITER_M(ITER)] = ann_calc_weights(FEATURE_LENGTH, PARAMETERS.HIDDEN_LAYER_LENGTH, NUM_CLASSES, NORMALIZED.TRAINING, PARAMETERS.USE_MOMENTUM_TERM);
    else
        [INP2HID_NEW, HID2OUT_NEW, ITER_M(ITER)] = ann_calc_weights_hess(FEATURE_LENGTH, PARAMETERS.HIDDEN_LAYER_LENGTH, NUM_CLASSES, NORMALIZED.TRAINING);
    end
    
    % 2. Classify trainings data
    % later on we will choose that classifier that best separates the training set
    [VAL, IND] = max(ann_calc_output(INP2HID_NEW, HID2OUT_NEW, NORMALIZED.TRAINING(:, 1:FEATURE_LENGTH)'));
    EVAL_MATRIX_TRAINING_NEW = [CLASSES(IND') DATA.TRAINING(:, end)];

    % 3. calculate probability of correct classification
    P_CORRECT_NEW = sum(EVAL_MATRIX_TRAINING_NEW(:, 1) == EVAL_MATRIX_TRAINING_NEW(:, 2))/TRAINING_LENGTH
    if P_CORRECT_NEW > P_CORRECT
        INP2HID   = INP2HID_NEW;
        HID2OUT   = HID2OUT_NEW;
        P_CORRECT = P_CORRECT_NEW;
    end
end



%% 3. CLASSIFICATIONPHASE
% ITER_M contains the number of iterations the backprop algorithm needed to
% converge
mean(ITER_M)
std(ITER_M)

[VAL, IND] = max(ann_calc_output(INP2HID, HID2OUT, NORMALIZED.CLASSIFICATION(:, 1:FEATURE_LENGTH)'));
EVAL_MATRIX = [CLASSES(IND') DATA.CLASSIFICATION(:, end)];