function [EVAL_MATRIX, PROB_EVAL_MATRIX] = bayesian_classifier(DATA, PARAMETERS)
%BAYESIAN_CLASSIFIER - this function realizes a bayesian classifier
% the loglikelihood pdfs are modelled by multivariate gaussian variables
% if the covariance matrix is singular, the classification rate is set to
% zero
% It returns a matrix EVAL_MATRIX which contains in the first column
% the estimated classes and in the second one the original classes.
%
% Additionally it returns a PROB_EVAL_MATRIX where each row contains the
% probabilities for one data sample of the classification set.
% 
% Possible parameters are:
%
% PARAMETERS.COVARIANCE_MATRIX  - 'full': estimate full covariance matrix
%                                 'diagonal': approximate covariance matrix as
%                                             diagonal matrix (all features are uncorrelated)
%                                 'spherical': assume C = sigma^2*I (spherical model)
%                                 Note: the covariance matrix is computed for each class

%% 1. INITIALISATION
[TRAINING_LENGTH, CLASSIFICATION_LENGTH, NUM_CLASSES, CLASSES, FEATURE_LENGTH] = get_lengths(DATA, PARAMETERS);

% type of covariance matrix
if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'COVARIANCE_MATRIX') == 0
    PARAMETERS.COVARIANCE_MATRIX = 'full';
end

% initialize evaluation matrix
EVAL_MATRIX = [NaN(CLASSIFICATION_LENGTH, 1) DATA.CLASSIFICATION(:, end)];
PROB_EVAL_MATRIX = [];

%% 2. TRAININGSPHASE
% calculate mu and sigma^2 for all emotions
for CLASSES_C = 1:NUM_CLASSES
    INDICES = DATA.TRAINING(:, end) == CLASSES(CLASSES_C);
    THETA(CLASSES_C).MEAN_VEC = mean(DATA.TRAINING(INDICES, 1:FEATURE_LENGTH));
    switch PARAMETERS.COVARIANCE_MATRIX
        case 'full'
            THETA(CLASSES_C).COV_MAT  = cov(DATA.TRAINING(INDICES, 1:FEATURE_LENGTH));
        case 'diagonal'
            THETA(CLASSES_C).COV_MAT  = diag(var(DATA.TRAINING(INDICES, 1:FEATURE_LENGTH)));
        case 'spherical'
            THETA(CLASSES_C).COV_MAT  = mean(var(DATA.TRAINING(INDICES, 1:FEATURE_LENGTH)))*eye(FEATURE_LENGTH);
    end
    
    % check whether covariance matrix can be inverted or not
    if rcond(THETA(CLASSES_C).COV_MAT) < 1e-13
        fprintf('Warning: Correlation-Matrix near singular\n');
        fprintf('for class %d\n', CLASSES(CLASSES_C));
        return
    end
    THETA(CLASSES_C).A_PRIORI = sum(INDICES)/TRAINING_LENGTH;
end


%% 3. CLASSIFICATIONPHASE
A_POSTERIORI = zeros(CLASSIFICATION_LENGTH, NUM_CLASSES);
for CLASSES_C = 1:NUM_CLASSES
    DIFFS = DATA.CLASSIFICATION(:, 1:FEATURE_LENGTH) - repmat(THETA(CLASSES_C).MEAN_VEC, CLASSIFICATION_LENGTH, 1);
    C     = chol(THETA(CLASSES_C).COV_MAT);
    TMP   = DIFFS / C;
    A_POSTERIORI(:,CLASSES_C) = exp(-0.5*sum(TMP.*TMP,2)) / ((2*pi)^(FEATURE_LENGTH/2)*prod(diag(C))) * THETA(CLASSES_C).A_PRIORI;
end

[VAL, INDEX] = max(A_POSTERIORI,[],2);
EVAL_MATRIX(:,1) = CLASSES(INDEX);

if nargout == 2
    PROB_EVAL_MATRIX = A_POSTERIORI./repmat(sum(A_POSTERIORI,2),1,NUM_CLASSES);
end

