function [BEST_MIX_PROBS, BEST_MEANS, BEST_COV_MAT, BEST_LL, AIC, AICC] = expectation_maximization_algorithm(SAMPLES, GMM_DIM)
%EXPECTATION_MAXIMIZATION_ALGORITHM - uses the EM algorithm to compute the
% unknown parameters of a mixture of gaussians. NUMBER_OF_MODES gives the
% number of pobulations
%
% see Mixture Densities, Maximum Likelihood and the EM Algorithm by
% Richard A. Redner and Homer F. Walker for the used equations
%
% it also calculates the LogLikelihood for the this model
% this can be used for the AIC or AICC
% AICC is the corrected version of the AIC and should be used in the case of small training set sizes
% (see the book of Burnham)

FEATURE_DIM       = size(SAMPLES, 2);
NUMBER_OF_SAMPLES = size(SAMPLES, 1);


if GMM_DIM ~= 1
    TOTAL_RUNS = 20;
    BEST_LL    = -inf;


    for RUNS = 1:TOTAL_RUNS

        %% SET INITIAL VALUES
        % we have two possibilities:
        % 1. use kmeans algorithm of the statistical toolbox to obtain initial
        %    values
        % 2. MIX_PROBS: mixing probabilities are set to a uniform distribution
        %    MEANS    : choose randomly GMM_DIM samples of SAMPLES and take them as initial values.
        %    COV_MAT  : set the covariance matrices to unity matrices

        warning off
        % we do only 5 iterations to avoid the same starting points of the EM algoritms.
        [IDX, INIT_MEANS] = kmeans(SAMPLES, GMM_DIM, 'emptyaction', 'singleton','maxiter',5);
        warning on
        MIX_PROBS = histc(IDX, 1:GMM_DIM)/NUMBER_OF_SAMPLES;
        MEANS = INIT_MEANS;
        for DIM = 1:GMM_DIM
            if length(find(IDX == DIM))/NUMBER_OF_SAMPLES < 0.2
                % sometimes one cluster consists only of one sample
                % then we need another initial value for the variance
                COV_MAT(:,:, DIM) =  cov(SAMPLES(IDX == DIM, :)) + 0.01*eye(FEATURE_DIM, FEATURE_DIM);
            else
                COV_MAT(:,:, DIM) = cov(SAMPLES(IDX == DIM, :));
            end
        end

        %% MAIN LOOP
        ITER = 1;
        LL_HIST = zeros(1,5);
        while (ITER < 6 || std(LL_HIST) > 1e-1)

            % check whether we converge to a valid covariance matrix
            % or whether covariance matrix gets singular
            LL = 0;
            for DIM = 1:GMM_DIM
                if (min(eig(COV_MAT(:,:,DIM))) < 1e-10)
                    MIX_PROBS = 1;
                    MEANS     = mean(SAMPLES);
                    COV_MAT(:,:,1) = cov(SAMPLES);
                    LL = -inf;  % we want a solution with a mixture of gaussians
                    break
                end
            end
            if LL == -inf
                break
            end

            PDF_VAL = zeros(NUMBER_OF_SAMPLES, GMM_DIM);
            for DIM = 1:GMM_DIM
                PDF_VAL(:,DIM) = multivariatenormpdf2(SAMPLES, MEANS(DIM,:), COV_MAT(:,:,DIM));
            end

            % calculate new values
            PDF_RATIO     = PDF_VAL ./ (PDF_VAL * MIX_PROBS*ones(1, GMM_DIM));
            SUM_PDF_RATIO = sum(PDF_RATIO)';
            NEW_MIX_PROBS = (SUM_PDF_RATIO .* MIX_PROBS)/ NUMBER_OF_SAMPLES;
            NEW_MEANS     = PDF_RATIO' * SAMPLES ./ (SUM_PDF_RATIO*ones(1, FEATURE_DIM));

            % after we know the new means, we can calculate the covariance matrices
            for DIM = 1:GMM_DIM
                % (4.9)
                %TMP = (SAMPLES - repmat(NEW_MEANS(DIM,:), NUMBER_OF_SAMPLES, 1)) .* repmat(sqrt(PDF_RATIO(:, DIM)),1,FEATURE_DIM);
                TMP = (SAMPLES - ones(NUMBER_OF_SAMPLES,1)*NEW_MEANS(DIM,:)) .* (sqrt(PDF_RATIO(:, DIM))*ones(1,FEATURE_DIM));
                NEW_COV_MAT(:,:,DIM) = TMP'*TMP / SUM_PDF_RATIO(DIM);
            end


            LL = sum(log(PDF_VAL * MIX_PROBS));
            LL_HIST = [LL_HIST(2:end) LL];

            MIX_PROBS = NEW_MIX_PROBS;
            MEANS     = NEW_MEANS;
            COV_MAT   = NEW_COV_MAT;
            ITER = ITER + 1;

            % break if more than 200 Iterations
            if ITER > 200
                fprintf('break because more than 200 Iterations\n');
                break
            end
        end
        if BEST_LL < LL
            BEST_MIX_PROBS = MIX_PROBS;
            BEST_MEANS     = MEANS;
            BEST_COV_MAT   = COV_MAT;
            BEST_LL        = LL;
        end
    end
end

% falls GMM_DIM == 1 oder EM-Algorithmus abgebrochen wurde, dann
% Verteilung mithilfe von einmodaler Gaussverteilung modelieren.
if exist('BEST_MIX_PROBS','var') == 0
    BEST_MIX_PROBS = 1;
    BEST_MEANS = mean(SAMPLES);
    BEST_COV_MAT(:,:,1) = cov(SAMPLES);
    BEST_LL = sum(log(multivariatenormpdf2(SAMPLES, BEST_MEANS, BEST_COV_MAT(:,:,1))));
end

% calculate Akaike's Information Criterion

U_GMM_DIM = length(BEST_MIX_PROBS);
% K =  NUMBER_PARAMS_IN_MIX_PROBS + NUMBER_PARAMS_IN_MEANS       + NUMBER_PARAMS_IN_COV_MATS
K   =  U_GMM_DIM - 1              + U_GMM_DIM*size(BEST_MEANS,2) + U_GMM_DIM*size(BEST_COV_MAT,1)*(size(BEST_COV_MAT,1)+1)/2;

if NUMBER_OF_SAMPLES > K+1
    AIC  = -BEST_LL + K;
    AICC = -BEST_LL + K + K*(K+1)/(NUMBER_OF_SAMPLES-K-1);
else
    AIC  = inf;
    AICC = inf;
end

%% SUBFUNCTION multivariatenormpdf2
function Y = multivariatenormpdf2(X, MU, COV_MAT)
% returns the value of the pdf at the positions X
% expects X to be a matrix where each row gives a point
DIMENSION = length(MU);
DIFFS = X - ones(size(X,1),1)*MU;  %is faster than DIFFS = X - repmat(MU, size(X,1), 1);
C     = chol(COV_MAT);
TMP   = DIFFS / C;
Y = exp(-0.5*sum(TMP.*TMP,2)) / ((2*pi)^(DIMENSION/2)*prod(diag(C)));