function [EVAL_MATRIX, PROB_EVAL_MATRIX] = gmm_classifier(DATA, PARAMETERS)
%GMM_CLASSIFIER - this function realizes a bayesian classifier
% the loglikelihood pdfs are modelled by multivariate gaussian mixture
% pdfs
% if the covariance matrix is singular, the classification rate is set to
% zero
% It returns a matrix EVAL_MATRIX which contains in the first column
% the estimated classes and in the second one the original classes.
% Additionally it returns a PROB_EVAL_MATRIX where each row contains the
% probabilities for one data sample of the classification set.
%
% Possible parameters are:
%
% PARAMETERS.GMM_DIM            - Dimension of Gaussian Mixture Model that is used
%                                 as pdf (if AICc should be used than set this variable
%                                 to 'NaN'

%% 1. INITIALISATION
[TRAINING_LENGTH, CLASSIFICATION_LENGTH, NUM_CLASSES, CLASSES, FEATURE_LENGTH] = get_lengths(DATA, PARAMETERS);

% initialize evaluation matrix
EVAL_MATRIX = [zeros(CLASSIFICATION_LENGTH, 1) DATA.CLASSIFICATION(:, end)];

% the default value for PARAMETERS.GMM_DIM is NaN which means that we
% use the AIC to determine the best representation of the pdf
if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'GMM_DIM') == 0
    PARAMETERS.GMM_DIM = NaN;
end


%% 2. TRAININGSPHASE
for CLASSES_C = 1:NUM_CLASSES
    INDICES = DATA.TRAINING(:, end) == CLASSES(CLASSES_C);

    if isnan(PARAMETERS.GMM_DIM) == 0
        [THETA(CLASSES_C).MIX_PROBS, THETA(CLASSES_C).MEANS, THETA(CLASSES_C).COV_MAT] = expectation_maximization_algorithm(DATA.TRAINING(INDICES,1:FEATURE_LENGTH), PARAMETERS.GMM_DIM);
    else
        for GMM_DIM = 1:4
            [THETA2(CLASSES_C,GMM_DIM).MIX_PROBS, THETA2(CLASSES_C,GMM_DIM).MEANS, THETA2(CLASSES_C,GMM_DIM).COV_MAT, LL(GMM_DIM), AIC(GMM_DIM), AICC(GMM_DIM)] = expectation_maximization_algorithm(DATA.TRAINING(INDICES,1:FEATURE_LENGTH), GMM_DIM);
        end
        % use AICC because of small sample size
        [VAL, IND] = min(AICC);
        THETA(CLASSES_C).MIX_PROBS = THETA2(CLASSES_C,IND).MIX_PROBS;
        THETA(CLASSES_C).MEANS     = THETA2(CLASSES_C,IND).MEANS;
        THETA(CLASSES_C).COV_MAT   = THETA2(CLASSES_C,IND).COV_MAT;
    end
    % check whether covariance matrices can be inverted or not
    for DIM = 1:length(THETA(CLASSES_C).MIX_PROBS)
        if rcond(THETA(CLASSES_C).COV_MAT(:,:,DIM)) < 1e-10
            fprintf('Warning: Matrix near singular\n');
            return
        end
    end

    % determine a priori probability for this class
    A_PRIORI(CLASSES_C) = sum(INDICES)/TRAINING_LENGTH;
end



%% 3. CLASSIFICATIONPHASE
A_POSTERIORI = zeros(CLASSIFICATION_LENGTH, NUM_CLASSES);
for CLASSES_C = 1:NUM_CLASSES
    TMP = zeros(CLASSIFICATION_LENGTH, length(THETA(CLASSES_C).MIX_PROBS));
    for DIM = 1:length(THETA(CLASSES_C).MIX_PROBS)
        TMP(:,DIM) = multivariatenormpdf2(DATA.CLASSIFICATION(:,1:FEATURE_LENGTH), THETA(CLASSES_C).MEANS(DIM,:), THETA(CLASSES_C).COV_MAT(:,:,DIM));
    end
    A_POSTERIORI(:,CLASSES_C) = (TMP * THETA(CLASSES_C).MIX_PROBS) * A_PRIORI(CLASSES_C);
end
[VAL, INDEX] = max(A_POSTERIORI,[],2);
EVAL_MATRIX(:,1) = CLASSES(INDEX);
PROB_EVAL_MATRIX = A_POSTERIORI ./ repmat(sum(A_POSTERIORI,2), 1, NUM_CLASSES);


%% SUBFUNCTION multivariatenormpdf2
function Y = multivariatenormpdf2(X, MU, COV_MAT)
% returns the value of the pdf at the positions X
% expects X to be a matrix where each row gives a point
DIMENSION = length(MU);
DIFFS = X - repmat(MU, size(X,1), 1);
C     = chol(COV_MAT);
TMP   = DIFFS / C;
Y = exp(-0.5*sum(TMP.*TMP,2)) / ((2*pi)^(DIMENSION/2)*prod(diag(C)));
