function [EVAL_MATRIX, PROB_EVAL_MATRIX] = naive_bayes_classifier(DATA, PARAMETERS)
%NAIVE_BAYES_CLASSIFIER - 
%
% It returns a matrix EVAL_MATRIX which contains in the first column
% the estimated classes and in the second one the original classes.
%
% Additionally it returns a PROB_EVAL_MATRIX where each row contains the
% probabilities for one data sample of the classification set.
% 
% Possible parameters are:
%
% PARAMETERS.FEATURE_DISTRIBUTION - 'gaussian','uniform'
%                                   'gaussian_kernel' -> parzen window method with
%                                                        gaussian kernel
%                                                        function
% PARAMETERS.USE_DECORRELATION    - if 1, then a whitening transform is
%                                   done, with 'gaussian', this should
%                                   yield the same classification rate as
%                                   the function 'bayes_classifier'

%% 1. INITIALISATION
[TRAINING_LENGTH, CLASSIFICATION_LENGTH, NUM_CLASSES, CLASSES, FEATURE_LENGTH] = get_lengths(DATA, PARAMETERS);

% feature distribution
if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'FEATURE_DISTRIBUTION') == 0
    PARAMETERS.FEATURE_DISTRIBUTION = 'gaussian';
end

% use decorrelation
if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'USE_DECORRELATION') == 0
    PARAMETERS.USE_DECORRELATION = 0;
end


% initialize evaluation matrix
EVAL_MATRIX = [zeros(CLASSIFICATION_LENGTH, 1) DATA.CLASSIFICATION(:, end)];


%% 2. TRAININGSPHASE

% calculate pdf-parameters for all emotions
for CLASSES_C = 1:NUM_CLASSES
    INDICES     = DATA.TRAINING(:, end) == CLASSES(CLASSES_C);
    NUM_INDICES = sum(INDICES);
    
    if PARAMETERS.USE_DECORRELATION
        [THETA(CLASSES_C).EIGEN_MATRIX,D] = eig(cov(DATA.TRAINING(INDICES, 1:FEATURE_LENGTH)));
        TRAINING_FEATURES = DATA.TRAINING(INDICES, 1:FEATURE_LENGTH) * THETA(CLASSES_C).EIGEN_MATRIX; 
        % we do not use the cholesky factorization, as
        % this yields different probabilities (because of possible
        % compaction/enlargement of the feature distribution) as it
        % is computed for every class independently. Instead we use a eigenvalue
        % decomposition (principal components)
    else
        TRAINING_FEATURES = DATA.TRAINING(INDICES, 1:FEATURE_LENGTH);
        
    end
    THETA(CLASSES_C).A_PRIORI = sum(INDICES)/TRAINING_LENGTH;

    switch PARAMETERS.FEATURE_DISTRIBUTION
        case 'gaussian'
            THETA(CLASSES_C).MEAN = sum(TRAINING_FEATURES) / NUM_INDICES;
            
            % we do not use var as this is slower than these lines
            DIFFS              = TRAINING_FEATURES - ones(NUM_INDICES,1) * THETA(CLASSES_C).MEAN; 
            THETA(CLASSES_C).VAR  = sum(DIFFS .* DIFFS) ./ (NUM_INDICES-1);
        case 'uniform'
            SORTED_VALUE = sort(TRAINING_FEATURES);
            THETA(CLASSES_C).MIN  = SORTED_VALUE(1,:);
            THETA(CLASSES_C).MAX  = SORTED_VALUE(end,:);
            THETA(CLASSES_C).PROB = 1 ./ (THETA(CLASSES_C).MAX' - THETA(CLASSES_C).MIN');
        case 'gaussian_kernel'
            THETA(CLASSES_C).POINTS = TRAINING_FEATURES;
    end
end


%% 3. CLASSIFICATIONPHASE
A_POSTERIORI = zeros(CLASSIFICATION_LENGTH, NUM_CLASSES);
for CLASSES_C = 1:NUM_CLASSES
    
    if PARAMETERS.USE_DECORRELATION
        CLASSIFICATION_FEATURES = DATA.CLASSIFICATION(:, 1:FEATURE_LENGTH) * THETA(CLASSES_C).EIGEN_MATRIX;
    else
        CLASSIFICATION_FEATURES = DATA.CLASSIFICATION(:, 1:FEATURE_LENGTH);
    end
    
    switch PARAMETERS.FEATURE_DISTRIBUTION
        case 'gaussian'
            DIFFS = CLASSIFICATION_FEATURES - ones(CLASSIFICATION_LENGTH,1) * THETA(CLASSES_C).MEAN;
            A_POSTERIORI(:,CLASSES_C) = exp(-0.5*( DIFFS.*DIFFS ) * (1./THETA(CLASSES_C).VAR')) ./ ((2*pi)^(FEATURE_LENGTH/2)*sqrt(prod(THETA(CLASSES_C).VAR))) * THETA(CLASSES_C).A_PRIORI;
        case 'uniform'
            % check whether feature is inside min/max
            MIN_MATRIX = CLASSIFICATION_FEATURES >= ones(CLASSIFICATION_LENGTH,1) * THETA(CLASSES_C).MIN;
            MAX_MATRIX = CLASSIFICATION_FEATURES <= ones(CLASSIFICATION_LENGTH,1) * THETA(CLASSES_C).MAX;
            A_POSTERIORI(:,CLASSES_C) = (MIN_MATRIX & MAX_MATRIX) * THETA(CLASSES_C).PROB;
        case 'gaussian_kernel'
            A_POSTERIORI(:,CLASSES_C) = calculate_kernel_density(CLASSIFICATION_FEATURES, THETA(CLASSES_C).POINTS);
    end
end

[VAL, INDEX] = max(A_POSTERIORI,[],2);
EVAL_MATRIX(:,1) = CLASSES(INDEX);

if nargout == 2
    PROB_EVAL_MATRIX = A_POSTERIORI./repmat(sum(A_POSTERIORI,2),1,NUM_CLASSES);
end

%% SUBFUNCTION kernel_density
function Y = calculate_kernel_density(X, POINTS)
% returns the value of the pdf at the position X when the
% pdf is approximated by gaussian kernels
NUM_POINTS   = size(POINTS,1);
NUM_FEATURES = size(POINTS,2);
NUM_PATTERNS = size(X, 1);
TMP = normpdf(repmat(X,[1 1 NUM_POINTS]),permute(repmat(POINTS,[1 1 NUM_PATTERNS]),[3 2 1]),1/sqrt(NUM_POINTS));

Y = NUM_FEATURES/NUM_POINTS * prod(sum(TMP,3),2);