function [EVAL_MATRIX, META_VECT] = evaluate_classifier(DATA, PARAMETERS)
%EVALUATE_CLASSIFIER - tests a given classifier on its performance
% DATA.FEATURES   - set of features which are used for training and classification
%
% Possible parameters are:
%
% PARAMETERS.CLASSIFIER         - contains the classifier that should be used for
%                                 the analysis
% PARAMETERS.TCR                - ratio between training patterns and all
%                                 available patterns. E.g. PARAMETERS.TCR = 0.5 corresponds
%                                 to the case that half of the patterns are
%                                 used for training and the other half for
%                                 classification.
%                                 If PARAMETERS.TCR = NaN then
%                                     use all patterns for training and classification.
%                                 If PARAMETERS.TCR = 0 then
%                                     use leave-one-out method
%                                 Otherwise use given ratio
% PARAMETERS.ITERATIONS         - number of iterations that are done to get
%                                 "statistical stable" results.
% PARAMETERS.FEATURE_REDUCTION  - 1: Fisher transformation (determinant criterion)
%                                 2: Fisher transformation (trace criterion)
%                                 3: PCA
%                                 NaN: no Transformation (Default)
% PARAMETERS.REDUCED_DIMENSION   - ignored if PARAMETERS.FEATURE_REDUCTION == NaN
%                                 = 1,... then reduce feature dimension (Default: 3)
% PARAMETERS.SHOW_STATUS         - if 1, then output little status

%% 1. INITIALISATION
if isfield(DATA,'FEATURES') == 0
    DATA.FEATURES = [DATA.TRAINING; DATA.CLASSIFICATION];
end

if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'CLASSIFIER') == 0
    PARAMETERS.CLASSIFIER = 'bayes_classifier';
end

if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'ITERATIONS') == 0
    if exist('PARAMETERS','var') && isfield(PARAMETERS,'TCR') && isnan(PARAMETERS.TCR)
        PARAMETERS.ITERATIONS = 1; % if TCR == NaN then only one iteration is needed
    else
        PARAMETERS.ITERATIONS = 20;
    end
end

if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'TCR') == 0
    PARAMETERS.TCR = 0.9;
end

if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'FEATURE_REDUCTION') == 0
    PARAMETERS.FEATURE_REDUCTION = NaN;
end

if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'REDUCED_DIMENSION') == 0
    PARAMETERS.REDUCED_DIMENSION = 3;
end

if exist('PARAMETERS','var') == 0 || isfield(PARAMETERS,'SHOW_STATUS') == 0
    PARAMETERS.SHOW_STATUS = 0;
end

NUM_DATA_SETS     = size(DATA.FEATURES,1);

if isnan(PARAMETERS.TCR) == 0
    if PARAMETERS.TCR == 0
        %leave-one-out method
        NUM_DATA_TRAINING = NUM_DATA_SETS-1;
        PARAMETERS.ITERATIONS = NUM_DATA_SETS;
    else
        NUM_DATA_TRAINING = round(NUM_DATA_SETS*PARAMETERS.TCR);
    end
    EVAL_MATRIX = zeros((NUM_DATA_SETS - NUM_DATA_TRAINING)*PARAMETERS.ITERATIONS,2);
    META_VECT   = zeros((NUM_DATA_SETS - NUM_DATA_TRAINING)*PARAMETERS.ITERATIONS,1);
else
    NUM_DATA_TRAINING = NUM_DATA_SETS;
    EVAL_MATRIX = zeros(NUM_DATA_SETS*PARAMETERS.ITERATIONS,2);
    META_VECT   = zeros(NUM_DATA_SETS*PARAMETERS.ITERATIONS,1);
end


%% 2. EVALUATE CLASSIFIER

for ITER = 1:PARAMETERS.ITERATIONS
    if PARAMETERS.SHOW_STATUS
        ITER
    end
    RANDPERM_M = randperm(NUM_DATA_SETS);
    if isnan(PARAMETERS.TCR) == 0
        if PARAMETERS.TCR == 0
            TMP_DATA.TRAINING         = DATA.FEATURES;
            TMP_DATA.TRAINING(ITER,:) = [];
            TMP_DATA.CLASSIFICATION   = DATA.FEATURES(ITER,:);
        else
            TMP_DATA.TRAINING       = DATA.FEATURES(RANDPERM_M(1:NUM_DATA_TRAINING), :);
            TMP_DATA.CLASSIFICATION = DATA.FEATURES(RANDPERM_M(NUM_DATA_TRAINING+1:end), :);
        end
    else
        TMP_DATA.TRAINING       = DATA.FEATURES;
        TMP_DATA.CLASSIFICATION = DATA.FEATURES;
    end
    if isnan(PARAMETERS.FEATURE_REDUCTION) == 0
        switch PARAMETERS.FEATURE_REDUCTION
            case 1 % Fisher transformation (determinant criterion)
                TMP_DATA = fisher_discriminant_analysis(TMP_DATA, PARAMETERS.REDUCED_DIMENSION);
                
            case 2 % Fisher transformation (trace criterion)
                TMP_DATA = fisher_discriminant_analysis_trace_criterion(TMP_DATA, PARAMETERS.REDUCED_DIMENSION);
                
            case 3 % PCA
                TMP_DATA = principal_component_analysis(TMP_DATA, PARAMETERS.REDUCED_DIMENSION);
        end
    end
    EVAL_STRING = sprintf('TMP = %s(TMP_DATA, PARAMETERS);',PARAMETERS.CLASSIFIER);
    eval(EVAL_STRING);
    if isnan(PARAMETERS.TCR) == 0
        EVAL_MATRIX(1+(NUM_DATA_SETS - NUM_DATA_TRAINING)*(ITER-1):(NUM_DATA_SETS - NUM_DATA_TRAINING)*ITER,:) = TMP;
        META_VECT(1+(NUM_DATA_SETS - NUM_DATA_TRAINING)*(ITER-1):(NUM_DATA_SETS - NUM_DATA_TRAINING)*ITER,:)   = TMP_DATA.CLASSIFICATION(:,end);
    else
        EVAL_MATRIX(1+NUM_DATA_SETS*(ITER-1):NUM_DATA_SETS*ITER,:) = TMP;
        META_VECT(1+NUM_DATA_SETS*(ITER-1):NUM_DATA_SETS*ITER,:)   = TMP_DATA.CLASSIFICATION(:,end);
    end
end